from PySide6.QtCore import QEvent
from PySide6.QtCore import Signal

from PySide6.QtGui import QPixmap
from PySide6.QtGui import QImage

from PySide6.QtWidgets import QWidget
from PySide6.QtWidgets import QLabel

from mimetica import Layer


class Thumbnail(QLabel):

    _selected = Signal(int)

    def __init__(
        self,
        index: int,
        layer: Layer,
        parent: QWidget = None,
        scale: int = 75,
        border_size: int = 2,
    ):
        super().__init__(parent)

        self.index = index
        self.layer = layer

        (height, width) = layer.image.shape
        self.border_size = border_size

        qimg = QImage(layer.path)
        pxm = QPixmap(qimg).scaledToHeight(scale - 2 * self.border_size)
        self.setPixmap(pxm)
        self.setFixedHeight(scale)
        self.setStyleSheet(f"border: {self.border_size}px solid #000000;")

    def mousePressEvent(
        self,
        event: QEvent,
    ):
        self._selected.emit(self.index)

    def deselect(self):
        self.setStyleSheet(f"border: {self.border_size}px solid #000000;")

    def select(self):
        self.setStyleSheet(f"border: {self.border_size}px solid #00aa11;")



import sys

from PySide6.QtWidgets import QMainWindow
from PySide6.QtWidgets import QApplication
from PySide6.QtWidgets import QFileDialog
from PySide6.QtWidgets import QTabWidget
from PySide6.QtGui import QAction
from PySide6.QtGui import QKeySequence
from PySide6.QtCore import Slot

import cloup
import platform
import multiprocessing as mp

from pathlib import Path

from mimetica import conf
from mimetica import logger
from mimetica import Tab


class MainWindow(QMainWindow):
    def __init__(self, *args, **kwargs):

        super().__init__()

        self.setWindowTitle("Mimetica")

        # Tab widget
        # ==================================================
        self.tabs = QTabWidget(self)
        self.setup_tabs()
        self.setCentralWidget(self.tabs)

        # Menu
        # ==================================================
        self.menu = self.menuBar()
        self.main_menu = self.menu.addMenu("File")
        self.setup_menu()

        # Window dimensions
        # ==================================================
        geometry = self.screen().availableGeometry()

        # Slots & signals
        # ==================================================

    def setup_tabs(self):
        self.tabs.setTabsClosable(True)
        self.tabs.tabCloseRequested.connect(self._remove_tab)

    def setup_menu(self):
        # Main menu
        # ==================================================

        # Open file
        act_open_file = QAction("Open file", self)
        act_open_file.triggered.connect(self.open_file)
        self.main_menu.addAction(act_open_file)

        # Open stack
        act_open_stack = QAction("Open stack", self)
        act_open_stack.triggered.connect(self.open_stack)
        self.main_menu.addAction(act_open_stack)

        # Exit
        act_exit = QAction("Exit", self)
        act_exit.setShortcut(QKeySequence.Quit)
        act_exit.triggered.connect(self.close)
        self.main_menu.addAction(act_exit)

    def open_file(
        self,
        file_name: Path = None,
    ):
        if not file_name:
            filter = [
                "BMP (*.bmp)",
                "PNG (*.png)",
                "TIFF (*.tif *.tiff)",
                "JPEG (*.jpg *.jpeg)",
            ]

            filter = ";;".join([f"*.{f}" for f in filter])

            file_name = QFileDialog.getOpenFileName(
                self,
                "Open file...",
                filter=filter,
            )[0]

        if file_name != "":
            self.open_stack([file_name])

    def open_stack(
        self,
        path: Path | None = None,
    ):
        if not path:
            path = QFileDialog.getExistingDirectory(self, "Open stack...")

        if path == "":
            return

        elif isinstance(path, (str, Path)):
            path = Path(path)
            if path.is_file():
                paths = [path]

            else:
                paths = []
                extensions = {
                    ".bmp",
                    ".tif",
                    ".tiff",
                    ".png",
                    ".jpg",
                    ".jpeg",
                }

                for file in path.iterdir():
                    if str(file).startswith("."):
                        continue

                    if file.suffix.lower() in extensions:
                        paths.append(file.resolve().absolute())

        elif isinstance(path, list):
            paths = [Path(p) for p in path]

        else:
            raise TypeError(f"Invalid path type: {type(path)}")

        if len(paths) != 0:
            self._add_tab(paths)

    @Slot(int)
    def _remove_tab(
        self,
        pos: int,
    ):
        if pos < self.tabs.count():
            self.tabs.removeTab(pos)

    def _add_tab(
        self,
        paths: list[Path],
    ):
        tab = Tab(paths)
        name = paths[0].parent.name
        idx = self.tabs.addTab(tab, name)
        self.tabs.setCurrentIndex(idx)


@cloup.command()
@cloup.option_group(
    "Input",
    "Open one or more images.",
    cloup.option(
        "-i",
        "--image",
        type=str,
        default=None,
        help="Open a single image.",
    ),
    cloup.option(
        "-s",
        "--stack",
        type=str,
        default=None,
        help="Open a stack (directory of images).",
    ),
    constraint=cloup.constraints.mutually_exclusive
)
def run(
    image: str | None,
    stack: str | None,
):

    # Set the multiprocessing context
    plt = platform.system()
    logger.warning(f"Running on {plt}")
    if plt.lower() == "windows":
        mp.set_start_method("spawn")
    else:
        mp.set_start_method("forkserver")

    # The main feature
    app = QApplication(sys.argv)
    mw = MainWindow()

    mw.showMaximized()
    mw.show()

    if image is not None:
        mw.open_file(image)
    elif stack is not None:
        mw.open_stack(stack)

    sys.exit(app.exec())



from .utils.config import conf
from .utils.logger import logger
from .gui.dock import Dock
from .gui.plot import Plot
from .scan.layer import Layer
from .gui.thumbnail import Thumbnail
from .scan.stack import Stack
from .gui.image import ImageView
from .gui.canvas import Canvas
from .gui.splitview import SplitView
from .gui.tab import Tab



import pyqtgraph as pg

from mimetica import Layer
from mimetica.gui.roi.line import RadialLine
from mimetica.gui.roi.contour import Contour


class ImageView(pg.ImageView):
    def __init__(
        self,
        *args,
        **kwargs,
    ):

        plot = pg.PlotItem()
        plot.setLabel(axis="left", text="Y-axis")
        plot.setLabel(axis="bottom", text="X-axis")
        kwargs["view"] = plot
        super().__init__(*args, **kwargs)

        self.ui.histogram.hide()
        self.ui.roiBtn.hide()
        self.ui.menuBtn.hide()

        self.radial_roi = None
        self.phase_roi = None

        self.view.invertX(False)
        self.view.invertY(False)

    def set_roi(
        self,
        layer: Layer,
    ):
        (cx, cy) = layer.centre[0], layer.centre[1]

        if self.radial_roi is not None:
            self.removeItem(self.radial_roi)
        self.radial_roi = Contour((cy, cx), radius=1)
        self.addItem(self.radial_roi)

        if self.phase_roi is not None:
            self.removeItem(self.phase_roi)
        self.phase_roi = RadialLine([(cy + 0.5, cx + 0.5), (cy + 1, cx + 0.5)])
        self.addItem(self.phase_roi)



from pathlib import Path

from PySide6.QtCore import Slot

import skimage as ski
import skimage.draw as skd

import shapely as shp

import numpy as np

from mimetica import conf
from mimetica import utils


class Layer:
    def __init__(
        self,
        path: Path,
    ):
        self.path = Path(path).resolve().absolute()

        image = ski.io.imread(str(self.path), as_gray=True)
        image = np.fliplr(image.T)

        # Pad the image to accommodate for the contour
        # ==================================================
        pad = 2
        self.image = np.pad(
            image,
            ((pad, pad), (pad, pad)),
            mode="constant",
            constant_values=0,
        )

        # Image properties
        # ==================================================
        # Minimal bounding circle
        self.mbc, self.mbr = utils.compute_minimal_bounding_circle(self.image)
        # Use np.roll to swap h and w for the centre.
        self.centre = np.roll(
            np.array(self.mbc.centroid.xy, dtype=np.uint32).flatten(), 1
        )
        self.radius = int(shp.minimum_bounding_radius(self.mbc))
        self.radial_range = np.empty([])
        self.radial_profile = np.empty([])
        self.phase_range = np.empty([])
        self.phase_profile = np.empty([])

        # Extract contour, centre, etc.
        # ==================================================
        self.process()

    def make_mask(self) -> np.ndarray:
        """
        Create a mask for this layer.

        Returns:
            A mask as a NumPy array.
        """
        Y, X = np.meshgrid[: self.image.shape[0], : self.image.shape[1]]
        mask = np.sqrt((X - self.centre[0]) ** 2 + (Y - self.centre[1]) ** 2)
        mask = np.exp(-3 * mask / mask.max())
        return mask

    def process(self):
        """
        Process this layer.
        For now, this is limited to computing the radial and phase profiles.
        """
        self.compute_radial_profile()
        self.compute_phase_profile()

    @Slot()
    def compute_radial_profile(self):
        # Create an empty array
        self.radial_profile = np.zeros((conf.radial_samples,))
        self.radial_range = np.linspace(0, 1, len(self.radial_profile))
        self.radii = np.linspace(1.0, self.radius, conf.radial_samples)

        for idx, radius in enumerate(self.radii):
            # Create a virtual circle with the right radius
            rr, cc = skd.circle_perimeter(
                self.centre[0], self.centre[1], int(radius), method="andres"
            )
            # Find out how much material is sampled by the circle
            # and compute the density
            circle = self.image[rr, cc]
            material = np.count_nonzero(circle)
            self.radial_profile[idx] = material / circle.size

    @Slot()
    def compute_phase_profile(self):
        # Coordinates of the central point
        (cx, cy) = self.centre

        # Get the coordinates of each pixel of the MBC
        # (cont_xs, cont_ys) = skd.circle_perimeter(cx, cy, self.radius)

        # Compute the angles from the pixel coordinates.
        # _cont_xs = cont_xs - cx
        # _cont_ys = cont_ys - cy
        # ratios = _cont_xs / np.sqrt(_cont_xs**2 + _cont_ys**2)
        # angles = np.arccos(ratios)
        # angles = np.where(_cont_ys < 0, 2 * np.pi - angles, angles)
        angles = conf.phase_samples

        # X and Y datasets for the phase profile
        self.phase_range = np.linspace(0, 360, angles, endpoint=False)
        self.phase_profile = np.zeros_like(self.phase_range)

        # Draw lines ('spokes') from the centre to the MBC, each rotated by
        # an angle determined by the number of phase samples.
        spokes = [
            shp.affinity.rotate(
                shp.LineString(
                    [
                        shp.Point(*self.centre),
                        shp.Point(self.centre[0], self.centre[1] + self.mbr + 1),
                    ],
                ),
                angle,
                origin=shp.Point(*self.centre),
            )
            for angle in range(angles)
        ]

        mbc_rr, mbc_cc = utils.draw_sorted_circle(self.centre, self.mbr)

        circle = shp.LineString(zip((mbc_rr).tolist(), (mbc_cc).tolist()))
        endpoints = [
            np.array(spoke.intersection(circle).coords.xy, dtype=np.uint32)
            .flatten()
            .tolist()
            for spoke in spokes
        ]

        # Compute the X and Y coordinates for the pixels
        # corresponding to the angles calculated above
        # end_xs = (self.radius * np.cos(angles) + cx).astype(np.int32)
        # end_ys = (self.radius * np.sin(angles) + cy).astype(np.int32)

        # for idx, (end_x, end_y) in enumerate(zip(cont_ys, cont_xs)):
        for idx, angle in enumerate(self.phase_range):
            (end_x, end_y) = endpoints[idx]
            # Create a virtual line from the centre to the contour
            rr, cc = np.array(skd.line(cx, cy, end_x, end_y))
            # Find out how much material is sampled by the line
            # and compute the density
            line = self.image[rr, cc]
            material = np.count_nonzero(line)
            self.phase_profile[idx] = material / line.size



from dataclasses import dataclass
from datetime import datetime
from PySide6.QtCore import QSettings
from PySide6.QtGui import QColor
from omegaconf import OmegaConf

@dataclass
class Conf(QSettings):
    WindowGeometry: str = "window/geometry"
    WindowPosition: str = "window/position"
    InactivePlotColour: str = "plots/inactive/colour"
    ShowInactivePlots: str = "plots/inactive/show"
    ActivePlotColour: str = "plots/active/colour"
    InactiveSliceColour: str = "slice/inactive/colour"
    ActiveSliceColour: str = "slice/active/colour"
    SliceContourColour: str = "slice/contour/colour"
    ShowStack: bool = "stack/show"
    RadialSamples: int = "analysis/radial_samples"
    PhaseSamples: int = "analysis/phase_samples"

    def __init__(self, *args, **kwargs):
        super().__init__("Mimetica", "Mimetica", *args, **kwargs)
        self.setValue("Copyright", f"{datetime.now().year} Alexander Hadjiivanov")

    @property
    def window_geometry(self):
        pass

    # Show inactive plots
    @property
    def show_inactive_plots(self) -> bool:
        return self.value(Conf.ShowInactivePlots, True, bool)

    @show_inactive_plots.setter
    def show_inactive_plots(
        self,
        value: QColor,
    ):
        self.setValue(Conf.ShowInactivePlots, value)

    # Inactive plot colour
    @property
    def inactive_plot_colour(self) -> QColor:
        return self.value(
            Conf.InactivePlotColour,
            QColor(147, 147, 147, 100),
            QColor,
        )

    @inactive_plot_colour.setter
    def inactive_plot_colour(
        self,
        value: QColor,
    ):
        self.setValue(Conf.InactivePlotColour, value)

    # Active plot colour
    @property
    def active_plot_colour(self) -> QColor:
        return self.value(
            Conf.ActivePlotColour,
            QColor(51, 255, 51, 255),
            QColor,
        )

    @active_plot_colour.setter
    def active_plot_colour(
        self,
        value: QColor,
    ):
        self.setValue(Conf.ActivePlotColour, value)

    # Slice contour colour
    @property
    def slice_contour_colour(self) -> QColor:
        return self.value(
            Conf.SliceContourColour,
            QColor(255, 255, 0, 255),
            QColor,
        )

    @slice_contour_colour.setter
    def slice_contour_colour(
        self,
        value: QColor,
    ):
        self.setValue(Conf.SliceContourColour, value)

    # Show stack
    @property
    def show_stack(self) -> bool:
        return self.value(Conf.ShowStack, True, bool)

    @show_stack.setter
    def show_stack(
        self,
        value: bool,
    ):
        self.setValue(Conf.ShowStack, value)

    # Radial segments
    @property
    def radial_samples(self) -> int:
        return self.value(Conf.RadialSamples, 200, int)

    @radial_samples.setter
    def radial_samples(
        self,
        value: int,
    ):
        self.setValue(Conf.RadialSamples, value)

    # Phase segments
    @property
    def phase_samples(self) -> int:
        return self.value(Conf.PhaseSamples, 360, int)

    @phase_samples.setter
    def phase_samples(
        self,
        value: int,
    ):
        self.setValue(Conf.PhaseSamples, value)


conf = Conf()
